package org.oauth.scribe;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Scanner;

import org.scribe.builder.ServiceBuilder;
import org.scribe.builder.api.FacebookApi;
import org.scribe.builder.api.GoogleApi;
import org.scribe.builder.api.LinkedInApi;
import org.scribe.builder.api.TwitterApi;
import org.scribe.builder.api.YahooApi;
import org.scribe.model.Token;
import org.scribe.model.Verifier;
import org.scribe.oauth.OAuthService;

public class ScribeHelp {

	private static OAuthService service;
	public static final String FACEBOOK = "facebook";
	public static final String LINKEDIN = "linkedin";
	public static final String GOOGLE = "google";
	public static final String TWITTER = "twitter";
	public static final String YAHOO = "yahoo";
	protected  static final Token EMPTY_TOKEN = null;
	
	
	public static OAuthService getService(String source, String api_key, String key_secret) {
		ServiceBuilder temp = new ServiceBuilder();
		//OAuthService temp = new ServiceBuilder().provider(LinkedInApi.class);
		switch(source) {
			case FACEBOOK:
				temp = temp.provider(FacebookApi.class);
				break;
			case LINKEDIN:
				temp = temp.provider(LinkedInApi.class);
				break;
			case GOOGLE:
				temp = temp.provider(GoogleApi.class);
				break;
			case TWITTER:
				temp = temp.provider(TwitterApi.class);
				break;
			case YAHOO:
				temp = temp.provider(YahooApi.class);
				break;
		}
		
		service = temp
				.apiKey(api_key)
				.apiSecret(key_secret)
				.build();
		
		return service;
	}
	
	public static Token getAccessToken(String source, OAuthService srv) {
		service = srv;
		Token token = new Token("","");
		try {
			FileInputStream fis = new FileInputStream(source + ".dat");
			ObjectInputStream ois = new ObjectInputStream(fis);
			
			token = (Token) ois.readObject();
		}
		catch (IOException ioe) {
			Scanner in = new Scanner(System.in);
			Verifier verifier;
			Token requestToken;
			String authorizationUrl;
			System.out.println("=== " + source + "'s OAuth Workflow ===");
			if (FACEBOOK.equals(source)) {
				    System.out.println();
				    // Obtain the Authorization URL
				    System.out.println("Fetching the Authorization URL...");
				    authorizationUrl = service.getAuthorizationUrl(EMPTY_TOKEN);
				    System.out.println("Got the Authorization URL!");
				    System.out.println("Now go and authorize Scribe here:");
				    System.out.println(authorizationUrl);
				    System.out.println("And paste the authorization code here");
				    System.out.print(">>");
				    verifier = new Verifier(in.nextLine());
				    System.out.println();
				    
				    // Trade the Request Token and Verfier for the Access Token
				    System.out.println("Trading the Request Token for an Access Token...");
				    token = service.getAccessToken(EMPTY_TOKEN, verifier);
			}
			else if (GOOGLE.equals(source)) {
				    // Obtain the Request Token
				    requestToken = service.getRequestToken();

				    System.out.println("Now go and authorize Scribe here:");
				    System.out.println("https://www.google.com/accounts/OAuthAuthorizeToken?oauth_token=" + requestToken.getToken());
				    System.out.println("And paste the verifier here");
				    System.out.print(">>");
				    verifier = new Verifier(in.nextLine());
				    System.out.println();

				    // Trade the Request Token and Verfier for the Access Token
				    System.out.println("Trading the Request Token for an Access Token...");
				    token = service.getAccessToken(requestToken, verifier);
			}
			else if (YAHOO.equals(source) || TWITTER.equals(source) || LINKEDIN.equals(source)) {
				    // Obtain the Request Token;
				    requestToken = service.getRequestToken();
				    
				    System.out.println("Go and authorize Scribe here:");
				    System.out.println(service.getAuthorizationUrl(requestToken));
				    System.out.println("And paste the verifier here");
				    System.out.print(">>");
				    verifier = new Verifier(in.nextLine());
				    System.out.println();
				
				    // Trade the Request Token and Verfier for the Access Token
				    token = service.getAccessToken(requestToken, verifier);
			}

		    
		    // Save the access token
		    try {
			    FileOutputStream fos = new FileOutputStream(source + ".dat");
				ObjectOutputStream oos = new ObjectOutputStream(fos);
				oos.writeObject(token);
		    }
		    catch (Exception ex) {
		    	ex.printStackTrace();
		    }
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		
		
		return token;
	}

}
