package org.semanticdatacapture;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.input.SAXBuilder;
import org.jdom2.output.Format;
import org.jdom2.output.XMLOutputter;
import org.plugin.startclass.SearchPlugin;

public class DataCapture {
	
	static Document doc, temp;
	static ArrayList<HashMap<String,String>> inputParameters, resultParameters;
	static ArrayList<String> parameters, parametersIn, parametersOut;
	static Element currentDataItem;
	static SearchPlugin currentPlugin;
	static HashMap<String, String> inputCandidate;
	static ArrayList<SearchPlugin> plugins;
	
	public static void main (String[] args) {
		long start, stop, startP, stopP, sum = 0, sumP = 0, startup = 0;
		start = System.currentTimeMillis();
		fillPlugins();
		inputParameters = new ArrayList<HashMap<String,String>>();
		boolean repeat = true;
		
		// Initialize the starting XML file
		doc = new Document();
		//Document temp;
		Element root = new Element("data");
		root.addContent(new Element("dataItem"));
		//doc.addContent(root);
		doc.setRootElement(root);

		//test variables
		int j = 0;
		
		stop = System.currentTimeMillis();
		startup = stop - start;
		
		while (repeat) {
			repeat = false;
			System.out.println("Iteration number: " + (j++));
			System.out.println();
			for (int i = 0; i < plugins.size(); i++) {
				System.out.println("Plugin: " + i);
				currentPlugin = plugins.get(i);
				if (getInputParametersForPlugin()) {
					//System.out.println(inputParameters);
					startP = System.currentTimeMillis();
					resultParameters = currentPlugin.makeQuery(inputParameters);
					stopP = System.currentTimeMillis();
					sumP = sumP + stopP - startP;
					//System.out.println(resultParameters);
					temp = doc.clone();
					insertDataToXML();
					if (!equalDocs())
						repeat = true;
				}
			}
		}
		printDoc();
		stop = System.currentTimeMillis();
		sum = stop - start;
		System.out.println("Aplikacija: " + sum + " ms");
		System.out.println("Plugini: " + sumP + " ms");
		System.out.println("Startup: " + startup + " ms");
	}
	
	
	private static void fillPlugins() {
		SAXBuilder sb = new SAXBuilder();
		Class c;
		HashMap<String, String> data = new HashMap<String, String>();
		plugins = new ArrayList<SearchPlugin>();
		try {
			Document loginData = sb.build(new File("plugins.xml"));
			List<Element> pluginList = loginData.getRootElement().getChildren();
			for (Element child : pluginList) {
				if (child.getText() != null) {
					c= Class.forName(child.getTextTrim());
					plugins.add((SearchPlugin)c.newInstance());
				}

			}
		}
		
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	
	/*
	 * Compares the old document and the document after inserting data.
	 */
	private static boolean equalDocs() {
		Element root1 = temp.getRootElement();
		Element root2 = doc.getRootElement();
		// Return true if root1 contains every element in root2! The other way isn't necessary to check because we never delete elements.
		
		return travelThroughTree(root1, root2);
	}
	
	private static boolean travelThroughTree (Element root1, Element root2) {
		List<Element> children1, children2;
		Element child1, child2;
		boolean foundMatch = false;
		children2 = root2.getChildren();
		
		// If it's not a leaf:
		if (children2.size() > 0) {
			for (int j = 0; j < children2.size(); j++) {
			//for (Element child2 : root2.getChildren()) {
				child2 = children2.get(j);
				foundMatch = false;
				children1 = root1.getChildren(child2.getName());
				for (int i = 0; i < children1.size() && !foundMatch; i++) {
					child1 = children1.get(i);
					if (travelThroughTree(child1, child2)) {
						foundMatch = true;
					}
				}
				if (!foundMatch) {
					return false;
				}
			}
		}
		else {
			if (root1.getText() != null && root2.getText() != null && root1.getText().equals(root2.getText())) {
				return true;
			}
			else {
				return false;
			}
		}

		return true;
	}
	
	private static void printDoc () {
		try{
			XMLOutputter outputter = new XMLOutputter(Format.getPrettyFormat());
			String xmlString = outputter.outputString(doc);

			System.out.println(xmlString);
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	private static boolean insertDataToXML() {
		boolean change = false;
		for (HashMap<String, String> data : resultParameters) {
			if (addDataToXML(data)) {
				change = true;
			}
		}
		
		return change;
	}
	
	private static boolean addDataToXML(HashMap<String,String> data) {
		//Element input = currentPlugin.inputXML();
		Element resultXML;
		boolean temp, change = false;
		
		// HashMap data must be added the input data!
		
		
		// Generate the resultXML from the HashMap data.
		resultXML = generateResultElementInputFirst(data);
		List<Element> candidates = getCandidates(doc, resultXML);
		
		if (candidates.size() == 0) {
			doc.getRootElement().addContent(resultXML);
			change = true;
		}
		
		for (Element candidate : candidates) {
			temp = mergeData(candidate, resultXML);
			if (temp)
				change = true;
		}
		
		return change;
	}
	
	
	/*
	 * Searches for legitimate candidates.
	 */
	private static ArrayList<Element> getCandidates(Document doc, Element resultXML) {
		ArrayList<Element> candidates = new ArrayList<Element>();
		List<Element> dataItems = doc.getRootElement().getChildren("dataItem");
		
		for (int i = 0; i < dataItems.size(); i++) {
			if (checkEquality(currentPlugin.inputXML(), dataItems.get(i), resultXML)) {
				candidates.add(dataItems.get(i));
			}
		}
		
		return candidates;
	}
	
	private static boolean checkEquality (Element input, Element dataItem, Element result) {
		List<Element> children = input.getChildren();
		Element child;
		boolean branchCheck = true;
		
		if (dataItem == null || result == null) {
			return false;
		}
		else if (children.size() == 0) {
			// You are in a leaf.
			if ( dataItem.getText() == null && result.getText() == null || dataItem != null && dataItem.getText().equals(result.getText())) {
				return true;
			}
			
		}
		else {
			for (int i = 0; i < children.size() && branchCheck == true; i++) {
				child = children.get(i);
				branchCheck = false;
				for (Element childDataItem : dataItem.getChildren(child.getName())) {
					if ( checkEquality(child, childDataItem, result.getChild(child.getName())) ) {
						branchCheck = true;
					}
				}
			}
			if (branchCheck) {
				return true;
			}
		}
		return false;
	}
	

	private static boolean mergeData (Element dataItem, Element result) {
		List<Element>  childrenDataItem, childrenResult;
		Element child, childDataItem;
		String container = "container";
		boolean check, branchCheck = false;
		childrenResult = result.getChildren();
		
		if (childrenResult.size() == 0) {
			// You are in a leaf.
			if ( dataItem.getText() == null && result.getText() != null ) {
				dataItem.setText(result.getText().toUpperCase());
				return true;
			}
			else if (dataItem.getText() != null && result.getText() != null) {
				if (dataItem.getText().equals(result.getText())) {
					return true;
				}
				else {
					return false;
				}
			}
			/* test purpose */
			else {
				System.out.println(dataItem.getText() + " : " + result.getText());
				return false;
			}
		}
		else {
			for (int i = 0; i < childrenResult.size(); i++) {
				check = false;
				child = childrenResult.get(i);
				childrenDataItem = dataItem.getChildren(child.getName());
				branchCheck = false;
				for (int j = 0; j < childrenDataItem.size() && !branchCheck; j++) {
					// Only one branch should suffice the conditions if any!
					childDataItem = childrenDataItem.get(j);
					check = mergeData(childDataItem, child);
					if (check) {
						branchCheck = true;
					}
					if ( !check && !"true".equals(result.getAttributeValue(container))) {
						return false;
					}
				}
				if (!branchCheck ) {
					// The data is not added yet, so it has do be added now.
					dataItem.addContent(child.clone());
				}
			}
		}
		return true;
	}
	
	private static Element generateResultElementInputFirst(HashMap<String, String> data) {
		Element result = currentPlugin.inputXML();
		addOutputToInput(result, currentPlugin.outputXML());
		
		// Add data to skeleton
		addDataToResult(data, result);
		
		return result;
	}
	
	private static boolean getInputParametersForPlugin() {
		List<Element> dataItems = doc.getRootElement().getChildren();
		boolean querable = false;
		
		Element result = currentPlugin.outputXML();
		addOutputToInput(result, currentPlugin.inputXML());
		
		inputParameters.clear();
		parameters = new ArrayList<String>();
		parameters = getParameters(result, parameters);
		parametersIn = getParameters(currentPlugin.inputXML(), new ArrayList<String>());
		parametersOut = getParameters(currentPlugin.outputXML(), new ArrayList<String>());
		parametersOut.remove("type");
		
		for (int i = 0; i < dataItems.size(); i++) {
			currentDataItem = dataItems.get(i);
			searchInputData(new HashMap<String, String>(), result, currentDataItem);
		}
		if (inputParameters.size() > 0)
			querable = true;
		
		return querable;
	}
	
	/*
	 * Searches through input and dataItem for the data in the leafs.
	 */
	private static boolean searchInputData(HashMap<String,String> inputData, Element input, Element dataItem) {
		String container = "container";
		Element child, childDataItem;
		List<Element> childrenDataItem;
		List<Element> inputChildren = input.getChildren();
		Set<String> commonParameters = new HashSet<String>();
		
		if (inputChildren.size() == 0) {
			for (String param : parametersOut ) {
				if (param.equals(input.getName())) {
					return false;
				}
			}
			String value = dataItem.getText();
			if (value == null)
				value = "";
			inputData.put(input.getName(), value);
			if (containsAllTheDataNecesary(inputData) /*&& input.getName().equals(parameters[parameters.length-1])*/) {
				addInputData(inputData);
			}
			return true;
		}
		else {
			for (int i = 0; i < inputChildren.size(); i++) {
				child = inputChildren.get(i);
				childrenDataItem = dataItem.getChildren(child.getName());
				if (childrenDataItem.size() == 0) {
					if (containsAllTheDataNecesary(inputData)) {
						addInputData(inputData);
					}
				}
				else {
					boolean branch = false;
					if ("true".equals(input.getAttributeValue(container))) {
						// Save the found parameters so you can delete the rest before going in a child from here.
						for (String param : inputData.keySet()) {
							commonParameters.add(param);
						}
						
					}
					for (int j = 0; j < childrenDataItem.size(); j++) {
						childDataItem = childrenDataItem.get(j);
						boolean check;
						check = searchInputData(inputData, child, childDataItem);
						if (check) {
							branch = true;
						}
						else if (!check && !"true".equals(input.getAttributeValue(container))) {
							return false;
						}
						if ("true".equals(input.getAttributeValue(container))) {
							for (String param : inputData.keySet()) {
								if (!commonParameters.contains(param))
									inputData.remove(param);
							}
						}
					}
					if (!branch) {
						return false;
					}
				}
			}
		}
		return true;
	}
	
	private static void addOutputToInput(Element input, Element output) {
		Element childI;
		if (output == null)
			return;
		if (output != null && output.hasAttributes()) {
			input.setAttribute("container", "true");
		}
		if (input != null && input.hasAttributes()) {
			output.setAttribute("container", "true");
		}
		for (Element childO : output.getChildren()) {
			childI = input.getChild(childO.getName());
			if (childI == null) {
				childI = new Element(childO.getName());
				if (childO.getText() != null) {
					childI.setText(childO.getText());
				}
				if (childO.hasAttributes()) {
					childI.setAttribute("container", "true");
				}
				input.addContent(childI);
			}
			addOutputToInput(childI, childO);
		}
	}
	
	
	private static void addDataToResult(HashMap<String,String> data, Element elm) {
		for (Element child : elm.getChildren()) {
			addDataToResult(data, child);
		}
		if (data.containsKey(elm.getName())) {
			elm.setText(data.get(elm.getName()).toUpperCase());
		} 
	}
	
	private static void addInputData (HashMap<String, String> inputData) {
		//inputParameters.add(inputData.);
		boolean addSwitch = false;
		for (HashMap<String, String> data : inputParameters) {
			addSwitch = false;
			for (String key : parametersIn) {
				if (!data.get(key).equals(inputData.get(key))) {
					addSwitch = true;
				}
			}
			if (!addSwitch) {
				return;
			}
		}

		HashMap<String, String> inp = new HashMap<String, String>();
		inp.putAll(inputData);
		inputParameters.add(inp);
	}

	private static boolean containsAllTheDataNecesary(HashMap<String, String> data) {
		for (String paramName : parametersIn) {
			if (!data.containsKey(paramName))
				return false;
		}
		
		return true;
	}

	private static ArrayList<String> getParameters(Element input, ArrayList<String> params) {
		List<Element> children = input.getChildren();
		
		if (children.size() == 0 && !input.getName().equals("dataItem")) {
			params.add(input.getName());
		}
		else {
			for (Element child : children) {
				getParameters(child,params);
			}
		}
		
		return params;
	}

	/*
	 * Compares XMLs as strings. If a plugin didn't make a query the xml should stay as it was before the plugin was called.
	 */
	private static boolean areEqual (Document first, Document second){
		// TODO: compare two strings without the attributes!!!
		XMLOutputter outputter = new XMLOutputter(Format.getPrettyFormat());
		String firstString = outputter.outputString(first);
		String secondString = outputter.outputString(second);
		
		if (firstString.equals(secondString)) {
			return true;
		}
		
		return false;
	}

}
